package org.semanticweb.rdf.utilities;

import java.util.*;
import java.io.*;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

  /** Transform XML to RDF.
  
    Basic idea:

     - toplevel Document node

     - for each startElement, create initial Element:

     <Element rdf:about="newid">
       <name rdf:resource="uri+local"/>
       <attribute>
         <Attribute>
           <name rdf:resource="uri+local"/>
           <value>...</value>
         </Attribute>
       </attribute>
       ...
     </Element>

     (analogously, for processing instructions and character data:
      PI and Text nodes)

     - plus link structure for children:

     <rdf:Description rdf:about="id">
       <first rdf:resource="eid">
         or
       <child rdf:resource="eid">
         or
       <next rdf:resource="eid">
     </rdf:Description>

  */

public class XML2RDF extends DefaultHandler {

  static String PARSER_NAME = "org.apache.xerces.parsers.SAXParser";
  String SYSTEM_NS = "http://www.semanticweb.org/xml2rdf#";
  String ANON_NODE_PREFIX;
  String DEFAULT_NS = "http://www.semanticweb.org/example#";
  String DEFAULT_ATTRIBUTE_NS = "http://www.semanticweb.org/xml2rdf/attribute#";

  PrintWriter writer;
  Locator locator;

  int nodeCounter; // to generate unique ids
  Node superNode;
  Node previousSibling;

  XML2RDF() {
    this.writer = new PrintWriter(System.out);
  }

  XML2RDF(String namespace, String attributeNamespace, PrintWriter writer) {
    if (namespace != null)
      DEFAULT_NS = namespace;
    if (attributeNamespace != null)
      DEFAULT_ATTRIBUTE_NS = attributeNamespace;
    this.writer = writer;
  }


  /*inner*/ class Node {
    String id;
    Node superNode;
    Node(String id, Node superNode) { 
      this.id = id;
      this.superNode = superNode;
    }
    String getId() { return id; }
    Node getSuperNode() { return superNode; }
  }


  String genId() {
    return ANON_NODE_PREFIX + (nodeCounter++);
  }


  // DocumentHandler methods

  /** Document locator */
  public void setDocumentLocator(Locator locator) {
    this.locator = locator;
  }

  /** Start document. */
  public void startDocument() {
    nodeCounter = 0;
    // what is the correct representation of anonymous nodes?
    // probably use Jena's method in future release
    ANON_NODE_PREFIX = "urn:rdf:an" + new Date().getTime() + "-";
    // write RDF header
    println("<?xml version=\"1.0\"?>\n");
    println("<rdf:RDF");
    println("   xmlns:rdf =\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"");
    println("   xmlns =\"" + SYSTEM_NS + "\"");
    println(">\n");
    // document node
    String id = genId();
    println("<Document rdf:about=\"" + id + "\">");
    String publicId = locator.getPublicId();
    if (publicId != null)
      println("  <publicId rdf:resource=\"" + publicId + "\"/>");
    String systemId = locator.getSystemId();
    if (systemId != null)
      println("  <systemId rdf:resource=\"" + systemId + "\"/>");
    println("</Document>\n");
    superNode = new Node(id, null);
    previousSibling = null;
  }

  public void endDocument() {
    println("</rdf:RDF>\n");
  }

  /** Start element. */
  public void startElement(String uri, String local, String qname, Attributes attrs) {
    // create element
    String id = genId();
    String resource = getResource(uri, local, qname);
    println("<Element rdf:about=\"" + id + "\">");
    println("  <name rdf:resource=\"" + resource + "\"/>");
    handleAttributes(attrs);
    println("</Element>\n");
    link(id);
    // update node "stack"
    Node node = new Node(id, superNode);
    superNode = node;
    previousSibling = null;
  }

  void handleAttributes(Attributes attrs) {
    for (int i = 0; i < attrs.getLength(); i++) {
      String resource = getAttributeResource(attrs.getURI(i), attrs.getLocalName(i), attrs.getQName(i));
      String value = attrs.getValue(i);
      String type = attrs.getType(i);
      println("  <attribute>");
      println("    <Attribute>");
      println("      <name rdf:resource=\"" + resource + "\"/>");
      print  ("      <value>");
      printString(value);
      println("</value>");
      println("      <type>" + type + "</type>");
      println("    </Attribute>");
      println("  </attribute>");
    }
  }

  /** End element. */
  public void endElement(String uri, String local, String qname) {
    // "pop"
    previousSibling = superNode;
    superNode = superNode.getSuperNode();
  }

  /** Characters. */
  public void characters(char ch[], int start, int length) {
    String string = new String(ch, start, length);
    // hmmm, we shouldn't trim etc. ...
    string = string.trim();
    if (!string.equals("")) {
      String id = genId();
      println("<Text rdf:about=\"" + id + "\">");
      print("  <text>");
      printString(string);
      println("</text>");
      println("</Text>\n");
      link(id);
      // update node "stack"
      Node node = new Node(id, superNode);
      previousSibling = node;
    }
  }

  /** Precessing instructions */
  public void processingInstruction(String target, String data) {
    String id = genId();
    println("<PI rdf:about=\"" + id + "\">");
    println("  <target>" + target + "</target>");
    print  ("  <data>");
    printString(data);
    println("</data>");
    println("</PI>\n");
    link(id);
    // update node "stack"
    Node node = new Node(id, superNode);
    previousSibling = node;
  }


  // auxiliaries

  void link(String id) {
    // link from super node
    if (superNode != null) {
      println("<rdf:Description rdf:about=\"" + superNode.getId() + "\">");
      if (previousSibling == null) // first child
        println("  <first rdf:resource=\"" + id + "\"/>");
      println("  <child rdf:resource=\"" + id + "\"/>");
      println("</rdf:Description>\n");
    }
    // link from previous sibling
    if (previousSibling != null) {
      println("<rdf:Description rdf:about=\"" + previousSibling.getId() + "\">");
      println("  <next rdf:resource=\"" + id + "\"/>");
      println("</rdf:Description>\n");
    }
  }

  String getResource(String uri, String local, String qname) {
    // qname ???
    if (uri.equals(""))
      uri = DEFAULT_NS;
    return uri + local;
  }

  String getAttributeResource(String uri, String local, String qname) {
    // qname ???
    if (uri.equals(""))
      uri = DEFAULT_ATTRIBUTE_NS;
    return uri + local;
  }

  void print(String string) {
    writer.print(string);
  }

  void print(char c) {
    writer.print(c);
  }

  void println(String string) {
    writer.println(string);
  }

  void printString(String string) {
    // use standard entities (&amp; etc.):
    int l = string.length();
    for (int i = 0; i < l; i++) {
      int c = string.charAt(i);
      switch (c) {
        case '&': print("&amp;"); break;
        case '<': print("&lt;"); break;
        case '>': print("&gt;"); break;
        case '\'': print("&apos;"); break;
        case '"': print("&quot;"); break;
        default: print((char)c);
      }
    }
  }


  //
  // ErrorHandler methods
  //

  /** Warning. */
  public void warning(SAXParseException ex) {
      System.err.println("[Warning] "+
                         getLocationString(ex)+": "+
                         ex.getMessage());
  }

  /** Error. */
  public void error(SAXParseException ex) {
      System.err.println("[Error] "+
                         getLocationString(ex)+": "+
                         ex.getMessage());
  }

  /** Fatal error. */
  public void fatalError(SAXParseException ex) throws SAXException {
      System.err.println("[Fatal Error] "+
                         getLocationString(ex)+": "+
                         ex.getMessage());
      throw ex;
  }

  /** Returns a string of the location. */
  private String getLocationString(SAXParseException ex) {
      StringBuffer str = new StringBuffer();

      String systemId = ex.getSystemId();
      if (systemId != null) {
          int index = systemId.lastIndexOf('/');
          if (index != -1)
              systemId = systemId.substring(index + 1);
          str.append(systemId);
      }
      str.append(':');
      str.append(ex.getLineNumber());
      str.append(':');
      str.append(ex.getColumnNumber());

      return str.toString();

  }


  /** Main program entry point. */
  public static void main(String args[]) throws Exception {
    if (args.length == 0) {
      System.err.println("arguments expected: [-ns <namespace] [-ans <attribute-namespace>]");
      System.err.println("                    (<filename/uri>|-) [-o <outputfile>]");
      System.err.println("  -ns <ns>     : use this as default namespace for elements");
      System.err.println("  -ans <ans>   : use this as default namespace for attributes");
      System.err.println("  -            : read from standard input");
      System.err.println("  -o <out>     : write output to file (default: standard output)");
      System.exit(1);
    }
    String namespace = null;
    String attributeNamespace = null;
    String uri = null; // null means standard input
    String outputFilename = null; // null means standard output
    int pos = 0;
    if ((args.length >= pos+2) && args[pos].equalsIgnoreCase("-ns")) {
      namespace = args[pos+1];
      pos = pos+2;
    }
    if ((args.length >= pos+2) && args[pos].equalsIgnoreCase("-ans")) {
      attributeNamespace = args[pos+1];
      pos = pos+2;
    }
    if (args.length >= pos+1) {
      uri = args[pos];
      if (uri.equals("-")) uri = null;
      pos++;
    }
    if ((args.length >= pos+2) && args[pos].equalsIgnoreCase("-o")) {
      outputFilename = args[pos+1];
      pos = pos+2;
    }
    if (args.length >= pos+1)
      System.err.println("arguments ignored: " + args[pos] + " ...");
    // do it
    PrintWriter writer;
    if (outputFilename != null)
      writer = new PrintWriter(new FileWriter(outputFilename));
    else
      writer = new PrintWriter(System.out);
    XML2RDF xml2rdf = new XML2RDF(namespace, attributeNamespace, writer);
    XMLReader parser = (XMLReader)Class.forName(PARSER_NAME).newInstance();
    parser.setContentHandler(xml2rdf);
    parser.setErrorHandler(xml2rdf);
    // parser.setFeature(....) ???
    if (uri != null)
      parser.parse(uri);
    else
      parser.parse(new InputSource(System.in));
    writer.flush();
    if (outputFilename != null)
      writer.close();
  }

}

